#include <inttypes.h>
#include <glcd.h>
#include <lfsr.h>
#include <../font/nums.h>
#include <../font/f9x14.h>
#include <../font/f8x11.h>
#include <../font/f13x16.h>

#include <util/delay.h>
#include <avr/interrupt.h>

#define POWER_ON    1     // out (keep high for power on)
#define TURN_POWER_OFF      PORTF &= ~(1<<POWER_ON)

void set_backlight(unsigned int my_volume)
  {
    
#define T3_PRESCALER_BITS  ((0<<CS12)|(1<<CS11)|(1<<CS10))           // same speed

    TCCR3A = (0 << COM3A1) | (0 << COM3A0)  // 00=normal port  01=normal port   10=clear on compare match, set at TOP   11 set on compare match
           | (0 << COM3B1) | (0 << COM3B0)
           | (0 << COM3C1) | (0 << COM3C0)
           | (0 << WGM31) | (1 << WGM30);

    TCCR3B = (0 << ICNC3)       // input noise canceler
           | (0 << ICES3)       // input cature edge
           | (0 << WGM33)
           | (1 << WGM32)
           | T3_PRESCALER_BITS;
    OCR3A = 0;
    OCR3B = 0;
    OCR3C = 0;
    //          0..100                 0..256
    //my_volume = backlight_brightness * global_brightness0_256;     // now we have a range of 0...25600
	
    if (my_volume == 0)
      {
        TCCR3A &=  ~(1 << COM3C1);
        TCCR3A &=  ~(1 << COM3C0);
      }
    else
      {
        OCR3CH = 0;
        OCR3CL = my_volume;
        TCCR3A |=   (1 << COM3C1);     // 10=clear on compare match, set at top
        TCCR3A &=  ~(1 << COM3C0);
      }
  }

  
// four different lines
void testLines(uint8_t iteration) {

    for (uint8_t i = 0; i < iteration; i++) {
//      glcdClearScreen(0);                          // clear only on in clipping rect
      glcdSetColor(0, SCREEN_COLOR);
      glcdSetColor(1, BLACK);
      glcdSetColor(2, GOLD);
      glcdSetColor(3, RED);
      for (glcdCoord_t x = SCREEN_LEFT; x <= SCREEN_RIGHT; x++) {
        glcdSetFgColor(YELLOW);
        glcdLine(x, SCREEN_TOP, x, SCREEN_BOTTOM);
        glcdSetFgColor(GREEN);
        glcdLine(SCREEN_RIGHT - x, SCREEN_TOP, SCREEN_RIGHT - x, SCREEN_BOTTOM);
        glcdSetFgColor(BLUE);
        glcdLine(SCREEN_LEFT, SCREEN_TOP, x, SCREEN_BOTTOM);
        glcdSetFgColor(RED);
        glcdLine(SCREEN_RIGHT, SCREEN_BOTTOM, SCREEN_RIGHT - x + SCREEN_LEFT, SCREEN_TOP);
      }
    }
}


// 255 random lines
void testLines2(uint8_t iteration) {

    for (uint8_t i = 0; i < iteration; i++) {
//      glcdClearScreen(0);
      for (uint8_t j = 0; j < 255; j++) {
        glcdSetFgColor(lfsr(16));
        glcdLine(lfsr(7) +2, lfsr(7) +2, lfsr(7) +2, lfsr(7) +2);
      }
    }
}

// animated ellipse shape
void testEllipses(uint8_t iteration) {

   glcdColor_t c[7];

   c[0] = YELLOW;
   c[1] = RED;
   c[2] = BLUE;
   c[3] = GREEN;
   c[4] = YELLOW;
   c[5] = RED;
   c[6] = BLUE;

   int8_t i, j = 0, count, k = 1, d = 1;

   for (; iteration > 0; iteration--) {
     glcdSetBkColor(NONE);
     for (count = 0; count < 3; count++) {
       for (i = 67; i > 0; i -= k) {
         glcdSetFgColor(c[j]);
         glcdSetFgColor(c[j +1]);
         glcdEllipse(65, 65, i, 67 - i);
         glcdSetFgColor(c[j +2]);
         glcdEllipse(65, 65, 67 - i, i);
       }
       if (++j >= 4) {j = 0;}
     }
     for (count = 0; count < 3; count++) {
       for (i = 67; i > 0; i -= k) {
         glcdSetFgColor(c[j]);
         glcdSetBkColor(c[j +1]);
         glcdEllipse(65, 65, i, 67 - i);
         glcdSetBkColor(c[j +2]);
         glcdEllipse(65, 65, 67 - i, i);
       }
       if (++j >= 4) {j = 0;}
     }
     k += d;
     if ((k > 5) | (k == 1)) {d = -d;}
   }
}

// random ellipses and circles with random border-/filling color
void testEllipses2(uint8_t iteration) {

    for (; iteration > 0; iteration--) {
//      glcdClearScreen(0);
      for (uint8_t i = 0; i < 255; i++) {
        glcdSetColors(lfsr(16), lfsr(16));
        glcdEllipse(lfsr(7) +2, lfsr(7) +2, lfsr(6), lfsr(6));
      }
    }
}

// random rectangles with random border-/filling color
void testRects(uint8_t iteration) {
//uint8_t val;
//	glcdSelectFont(f9x14, 0);
//	glcdSetBkColor(NONE);
//	glcdSetFgColor(BLACK);
    for (; iteration > 0; iteration--) {
//      glcdClearScreen(0);
      for (uint8_t i = 0; i < 255; i++) {
        glcdSetColors(NONE /*lfsr(16)*/, lfsr(16));
        glcdRectangle(lfsr(7) +2, lfsr(7) +2, lfsr(7) +2, lfsr(7) +2);
//			glcdSetBkColor(WHITE);
//			glcdSetFgColor(BLACK);
//			glcdMoveTo(1,1);
//			val = i;
//			glcdPut(val/100+48);
//			val = val % 100;
//			glcdPut(val/10+48);
//			val = val % 10;
//S			glcdPut(val+48);
//			glcdWait(500);
      }
    }
}

// Rotate a rectangle form source orientation to dest orientation such that
// the rectangle is on display always on same location. Please remember the orientation
// of the display in this library is not relative and thus the clipping rect is relative
// to the current orientation set. This simplify normaly the work with different orientations
// on same code. Means the same source with painting can be directly used with different orientations
// without any changes on coordinates. But for this demonstration we have to take care about this,
// because we want always printing on same display locations with different orientations.
// Follow function is'nt included directly in the library, but could be sometimes important to know.

void glcdRotateRect(glcdRect_t *rect, uint8_t source, uint8_t dest) {

    if ((source & 3) != (dest & 3)) {

      glcdCoord_t x1=0, x2=0, y1=0, y2=0;

      switch (source & 3) {
        case 0: {
          x1 = (*rect).X1;
          y1 = (*rect).Y1;
          x2 = (*rect).X2;
          y2 = (*rect).Y2;
          break;
        }
        case 1: {
          x1 = 131 - (*rect).Y2;
          x2 = 131 - (*rect).Y1;
          y1 = (*rect).X1;
          y2 = (*rect).X2;
          break;
        }
        case 2: {
          x1 = 131 - (*rect).X2;
          x2 = 131 - (*rect).X1;
          y1 = (*rect).Y2 - 1;
          y2 = (*rect).Y1 - 1;
          break;
        }
        case 3: {
          x1 = (*rect).Y1;
          x2 = (*rect).Y2;
          y1 = 131 - (*rect).X2;
          y2 = 131 - (*rect).X1;
          break;
        }
      }

      switch (dest & 3) {
        case 0: {
          (*rect).X1 = x1;
          (*rect).Y1 = y1;
          (*rect).X2 = x2;
          (*rect).Y2 = y2;
          break;
        }
        case 1: {
          (*rect).X1 = y1;
          (*rect).X2 = y2;
          (*rect).Y1 = 131 - x2;
          (*rect).Y2 = 131 - x1;
          break;
        }
        case 2: {
          (*rect).X1 = 131 - x2;
          (*rect).X2 = 131 - x1;
          (*rect).Y1 = y2 + 1;
          (*rect).Y2 = y1 + 1;
          break;
        }
        case 3: {
          (*rect).X1 = 131 - y2;
          (*rect).X2 = 131 - y1;
          (*rect).Y1 = x1;
          (*rect).Y2 = x2;
          break;
        }
      }
    }
}

void glcdNewOrientation(uint8_t orientation) {

    glcdRotateRect(&glcd_Clip, glcd_MemCtrl, orientation);
    glcdSetOrientation(orientation);
}

// display random 4 color font symbols with tranparancy and random rotated by 0,90,180 or 270 degree
void testSymbols(uint8_t iteration) {

    uint8_t mem = glcd_MemCtrl;     // save glcd_MemCtrl, thus orientation in 2 LSB's

    glcd_Flags.Bits.AutoLineFeed = 0;

    glcdSelectFont(nums, 0);        // font is stored in FLASH, thus no need for own read callback
    glcdSetBkColor(NONE);
    for (; iteration > 0; iteration--) {
//      glcdClearScreen(0);
      for (uint8_t i = 0; i < 255; i++) {
        glcdNewOrientation(lfsr(2));
        glcdSetColor(1, lfsr(16));
        glcdSetColor(2, lfsr(16));
        glcdSetColor(3, lfsr(16));
        glcdMoveTo(lfsr(7), lfsr(7));
        glcdDrawChar(lfsr(4) % 11 + 32);
      }
    }
    glcdNewOrientation(mem);
}



// display random standard font symbols
void testFont(uint8_t iteration) {

	glcdNewOrientation(0);
    glcd_Flags.Bits.AutoLineFeed = 0;

    glcdSelectFont(f9x14, 0);
    glcdSetBkColor(WHITE);
    glcdSetFgColor(BLACK);
    for (; iteration > 0; iteration--) {
//      glcdClearScreen(0);
      for (uint8_t i = 0; i < 255; i++) {
        glcdMoveTo(lfsr(7), lfsr(7));
        glcdDrawChar(lfsr(128) % 96 + 32);
      }
    }
}



// print text centered horizontal and vertical into bounds r
// demontrate the use of glcdCharsWidth()
void glcdPrintCenter(glcdRect_t r, char *text, uint8_t InFlashStored) {

    r.X1 += (r.X2 - r.X1 - glcdCharsWidth(text, InFlashStored)) / 2;
    r.Y1 += (r.Y2 - r.Y1 - glcd_FontHeight) / 2;
    glcdMoveTo(r.X1, r.Y1);
    glcdPrint(text, InFlashStored);
}


// demontrate own font read callback
uint8_t myRead(glcdFontData_t index) {

    return(pgm_read_byte(f9x14 + index));
}

// paint on the four possible orientation, as shown the coordinates are always the same
void testOrientation(void) {

    uint8_t mem = glcd_MemCtrl;          // save glcd_MemCtrl
    glcdFlags_t flags = glcd_Flags;      // save glcd_Flags

    glcd_Flags.All = 0;                  // all bits off, no clipping, no fixed font and no autolinefeed

    glcdRect_t r;
    glcdSetRect(r, 1, 1, 64, 64);

    glcdClearScreen(0);
    glcdSelectFont(0, myRead);           // font is stored in FLASH, demontrate read callback

    glcdSetOrientation(0);
    glcdSetColors(BLACK, NONE);
    glcdRectangle(r.X1, r.Y1, r.X2, r.Y2);
    glcdLine(r.X1, r.Y1, r.X2, r.Y2);
    glcdCircle(r.X1, r.Y1, 20);
    glcdMoveTo(r.X1 + 2, r.Y1 + 2);
    glcdPrintCenter(r, PSTR("0 Grad"), 1);

    glcdSetOrientation(1);
    glcdSetColors(RED, NONE);
    glcdRectangle(r.X1, r.Y1, r.X2, r.Y2);
    glcdLine(r.X1, r.Y1, r.X2, r.Y2);
    glcdCircle(r.X1, r.Y1, 20);
    glcdMoveTo(r.X1 + 2, r.Y1 + 2);
    glcdPrintCenter(r, PSTR("90 Grad"), 1);

    glcdSetOrientation(2);
    glcdSetColors(GREEN, NONE);
    glcdRectangle(r.X1, r.Y1, r.X2, r.Y2);
    glcdLine(r.X1, r.Y1, r.X2, r.Y2);
    glcdCircle(r.X1, r.Y1, 20);
    glcdMoveTo(r.X1 + 2, r.Y1 + 2);
    glcdPrintCenter(r, PSTR("180 Grad"), 1);

    glcdSetOrientation(3);
    glcdSetColors(YELLOW, NONE);
    glcdRectangle(r.X1, r.Y1, r.X2, r.Y2);
    glcdLine(r.X1, r.Y1, r.X2, r.Y2);
    glcdCircle(r.X1, r.Y1, 20);
    glcdMoveTo(r.X1 + 2, r.Y1 + 2);
    glcdPrintCenter(r, PSTR("270 Grad"), 1);

    glcdWait(2000);       // 2 seconds
    glcd_Flags = flags;
    glcdSetOrientation(mem);
}

// paint own bitmaps
void testBitmaps(void) {

    uint8_t Bitmap[512];
    glcdCoord_t width = glcd_Clip.X2 - glcd_Clip.X1 +1;
    glcdCoord_t height = glcd_Clip.Y2 - glcd_Clip.Y1 +1;
    int16_t i,j;

    for (i = 0; i < 512; i++) {      // create a bitmap
      Bitmap[i] = i * 2;
    }

    glcdSetAddr(glcd_Clip.X1, glcd_Clip.Y1, glcd_Clip.X2, glcd_Clip.Y2);  // bounds of bitmap on screen

    for (i = 0; i < 256; i++) {
		glcdSetAddr(glcd_Clip.X1, glcd_Clip.Y1, glcd_Clip.X2, glcd_Clip.Y2);  // bounds of bitmap on screen
      for (j = 0; j < height; j++) {
        glcdDisplayBuffer(&Bitmap[i + j], width, 0);
      }
    }

}

void glcdRandomClipRect(void) {

    while (1) {
      glcdSetRect(glcd_Clip, lfsr(7) +2, lfsr(7) +2, lfsr(7), lfsr(7));
      if ((glcd_Clip.X1 < glcd_Clip.X2) & (glcd_Clip.Y1 < glcd_Clip.Y2)) {break;}
    }
}

uint16_t volatile tick;
#define T2_PRESCALER   64
#if   (T2_PRESCALER==1)
  #define T2_PRESCALER_BITS   ((0<<CS22)|(0<<CS21)|(1<<CS20))
#elif (T2_PRESCALER==8)
  #define T2_PRESCALER_BITS   ((0<<CS22)|(1<<CS21)|(0<<CS20))
#elif (T2_PRESCALER==32)
  #define T2_PRESCALER_BITS   ((0<<CS22)|(1<<CS21)|(1<<CS20))
#elif (T2_PRESCALER==64)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(0<<CS21)|(0<<CS20))
#elif (T2_PRESCALER==128)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(0<<CS21)|(1<<CS20))
#elif (T2_PRESCALER==256)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(1<<CS21)|(0<<CS20))
#elif (T2_PRESCALER==1024)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(1<<CS21)|(1<<CS20))
#else
  #error T2_PRESCALER is not defined or is not a supported value
#endif

void init_timer2(void)
  {
    TCCR2A = (0<<FOC2A)    // reserved set to 0
           | (1<<WGM20)    // Mode 3: Fast PWM (Top = 0xFF)
           | (0<<COM2A1)   // 1,0 Clear OC2 on Compare Match, set OC2 at TOP.
           | (0<<COM2A0)
           | (1<<WGM21)
           | T2_PRESCALER_BITS;

    TCNT2 = 0;

    TIMSK2 |= (1<<TOIE2);  // enable overflow interrupt

    tick = 0;
  }

ISR(TIMER2_OVF_vect)
  {
    // We get an Interrupt every 16384 clocks
    // Max-Time ~1073*1e6 clocks = 67sec @ 16MHz
	uint16_t temp = tick;
	if (temp!=-1)
		tick=temp+1;

  }  


uint32_t get_time(void) {
	uint32_t temp;
	cli();
	temp = ((uint32_t) tick<<8)+TCNT2;
//    if ((temp & 0xFF) < 250)
      {  // low counter value, may be we missed a tick)
        if (TIFR2 & (1<<TOV2))  // pending overflow interrupt not yet processed
          temp+=0x100L;
      }
//	temp = (uint32_t) tick;
	sei();
	return temp;
}


void clear_time(void) {
	cli();
	tick = 0;
	sei();
}


void benchmark(uint32_t *time) {
    uint8_t o = 0;

	lfsr_S = lfsrinit_S;			// Set the start condition for every run to the same value
    glcdSetOrientation(o);         // set orientation
    glcdClearScreen(1);              // clear screen, ignore clipping

	glcd_Flags.Bits.Clipping = 0;         // deactivate clipping
	glcdSetColor(0, BLACK);
	glcdSetColor(2, BLACK);
	glcdSetColor(3, NONE);
//	glcd_Flags.Bits.Clipping = 1;         // activate clipping
	
	clear_time();

	testRects(1);
	time[0] = get_time();
	
	testLines2(10);
	time[1] = get_time();
	
	testLines(1);
	time[2] = get_time();
	
	testEllipses(1);
	time[3] = get_time();
	
	testEllipses2(1);
	time[4] = get_time();
	
	testSymbols(1);
	time[5] = get_time();
	
	testFont(5);
	time[6] = get_time();
	
	glcdSetRect(glcd_Clip, 10, 10, 122, 122);
	testBitmaps();
	glcdSetRect(glcd_Clip, 0, 0, 131, 131);
	time[7] = get_time();
	
}


void show_results(uint32_t time_val[10]) {
	uint8_t i;
	uint32_t val;
	
	glcd_Flags.Bits.Clipping = 0;         // deactivate clipping
	glcdSelectFont(f9x14, 0);
	glcdSetBkColor(NONE);
	glcdSetFgColor(BLACK);
	glcdClearScreen(1);
	//glcdSetOrientation(0);         // set orientation
	glcdMoveTo(2,0);
	glcdPrint("Benchmark-Results\n", 0);
	glcdSelectFont(f8x11, 0);
	for (i=0; i<8; i++) {
		glcdMoveTo(i>=5?66:2, 15*(i%5)+22);
		glcdPut(i+48);
		glcdPrint(":",0);
		if (time_val[i] == -1) {
			glcdPrint("NA\n",0);
		} else {
			val = time_val[i] - (i==0?0:time_val[i-1]);
			glcdPut(val/10000000L+48);
			val = val % 10000000L;
			glcdPut(val/1000000L+48);
			val = val % 1000000L;
			glcdPut(val/100000L+48);
			val = val % 100000L;
			glcdPut(val/10000L+48);
			val = val % 10000L;
			glcdPut(val/1000L+48);
			val = val % 1000L;
			glcdPut(val/100L+48);
			val = val % 100L;
			glcdPut(val/10L+48);
			val = val % 10L;
			glcdPut(val+48);
			glcdPrint("\n",0);
		}
	}
	glcdMoveTo(2,105);
	glcdPrint("Total: ",0);
	val = time_val[7];
	glcdPut(val/10000000L+48);
	val = val % 10000000L;
	glcdPut(val/1000000L+48);
	val = val % 1000000L;
	glcdPut(val/100000L+48);
	val = val % 100000L;
	glcdPut(val/10000L+48);
	val = val % 10000L;
	glcdPut(val/1000L+48);
	val = val % 1000L;
	glcdPut(val/100L+48);
	val = val % 100L;
	glcdPut(val/10L+48);
	val = val % 10L;
	glcdPut(val+48);
}

int main(void) {
    
	uint32_t time_val[10];
	
	set_backlight(200);
	
#ifndef USE_AUTOINIT
    glcdDisplayInit();
#endif
	
    glcdDisplayOn();

    glcdClearScreen(1);

      //printCharAsHex()
    glcdSelectFont(f9x14, 0);
    glcdSetBkColor(NONE);
    glcdSetFgColor(BLACK);

    uint32_t deviceId = glcdDisplayRead(0x04, 24);
    glcdPrint("Id:", 0);
    int8_t ch;
    ch = (((uint8_t*)&deviceId)[3] >> 4) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    ch = (((uint8_t*)&deviceId)[3]) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);
    
    ch = (((uint8_t*)&deviceId)[2] >> 4) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    ch = (((uint8_t*)&deviceId)[2]) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    ch = (((uint8_t*)&deviceId)[1] >> 4) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    ch = (((uint8_t*)&deviceId)[1]) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    ch = (((uint8_t*)&deviceId)[0] >> 4) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    ch = (((uint8_t*)&deviceId)[0]) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    deviceId = glcdDisplayRead(9, 32);
    glcdPrint("\nId:", 0);
    ch = (((uint8_t*)&deviceId)[3] >> 4) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    ch = (((uint8_t*)&deviceId)[3]) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    
    ch = (((uint8_t*)&deviceId)[2] >> 4) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    ch = (((uint8_t*)&deviceId)[2]) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    ch = (((uint8_t*)&deviceId)[1] >> 4) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    ch = (((uint8_t*)&deviceId)[1]) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    ch = (((uint8_t*)&deviceId)[0] >> 4) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    ch = (((uint8_t*)&deviceId)[0]) & 0x0f;
    if (ch > 9) glcdPut(ch + 55); else glcdPut(ch+48);

    glcdPrint("\n\nHello World!", 0);
    _delay_ms(3000);

	init_timer2();
	sei();
	while(1) {
		benchmark(&time_val[0]);
		show_results(&time_val[0]);

		_delay_ms(20000);
		_delay_ms(20000);
		_delay_ms(20000);
		TURN_POWER_OFF;
	}
	
	
    uint8_t o = 0;
    #define cnt  5

    while (0) {

      glcdSetOrientation(o++);         // next orientation
      glcdClearScreen(1);              // clear screen, ignore clipping

      glcd_Flags.Bits.Clipping = 0;         // deactivate clipping
      glcdSetColor(0, BLACK);
      glcdSetColor(2, BLACK);
      glcdSetColor(3, NONE);
      glcdFrame(glcd_Clip.X1 -2, glcd_Clip.Y1 -2, glcd_Clip.X2 +2, glcd_Clip.Y2 +2);
      glcd_Flags.Bits.Clipping = 1;         // activate clipping

      glcdClearScreen(1);
      testRects(cnt);
      glcdClearScreen(1);
      testLines2(cnt);
      glcdClearScreen(1);
      testLines(cnt);
      glcdClearScreen(1);
      testEllipses(cnt);
      glcdClearScreen(1);
      testEllipses2(cnt);
      glcdClearScreen(1);
      testSymbols(cnt);
      glcdClearScreen(1);
      testBitmaps();

      glcdClearScreen(1);
      testOrientation();

      glcdRandomClipRect();
    }
}

